import tarfile
from typing import Generator, List

try:
    import ndjson
except ModuleNotFoundError:
    pass

from ayeaye.connectors.base import DataConnector
from ayeaye.pinnate import Pinnate


class FlowerpotEngine:
    """File-access Interface to 'Flowerpot' style ND-JSON tarballs."""

    def __init__(self, tar_file):
        self._tar_file = tar_file

    @staticmethod
    def from_filename(filename) -> 'LazyFlowerpotReader':
        tf = tarfile.open(filename, mode='r:gz')
        return FlowerpotEngine(tf)

    @staticmethod
    def from_file(file_object) -> 'LazyFlowerpotReader':
        tf = tarfile.open(fileobj=file_object, mode='r:gz')
        return FlowerpotEngine(tf)

    @property
    def file_names(self) -> List[str]:
        """Get all filenames contained in a flowerpot tarball"""
        file_names = [f.name for f in self._tar_file.getmembers() if f.isfile()]
        return file_names

    def _extract_ndjson_to_list(self, filename) -> List[object]:
        """Deserialize an entire ndjson file to a python list"""
        with self._tar_file.extractfile(filename) as f:
            byte_string = f.read()
            return FlowerpotEngine._deserialize_ndjson_string(byte_string)

    def file_handles(self) -> Generator[object, None, None]:
        """
        Generator yielding (file_name (str), file_handle) for all files in this flowerpot.
        It's polite for the user of this generator to .close() the file handle.
        """
        for file_name in self.file_names:
            fh = self._tar_file.extractfile(file_name)
            yield file_name, fh

    @staticmethod
    def _deserialize_ndjson_string(byte_string) -> List[object]:
        """
        Deserialize the contents of a newline-delimited JSON string to a list
        Args:
            byte_string: The NDJSON contents to be deserialized
        Returns:
            list: Each individual JSON entry deserialized as Python objects
        """
        utf8_string = str(byte_string, 'utf-8')
        content = ndjson.loads(utf8_string)
        return content

    def items(self, file_name=None) -> Generator[object, None, None]:
        """
        Args:
            file_name: (str or list of str) file(s) within flowerpot if not
                        given, all files will be used.
        Generator returning each json object in a Flowerpot (across all individual files)
        """
        if file_name:
            selected_files = file_name if isinstance(file_name, list) else [file_name, ]
        else:
            selected_files = self.file_names

        for file in selected_files:
            yield from self._items_in_file(file)

    def _items_in_file(self, file) -> object:
        """Generator returning each json object in an ndjson file in :class:`Pinnate` form."""
        for deserialized_json_object in self._extract_ndjson_to_list(file):
            yield Pinnate(data=deserialized_json_object)


class FlowerPotConnector(DataConnector):
    engine_type = 'flowerpot://'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._flowerpot = None

    def connect(self):
        if self._flowerpot is None:
            file_path = self.engine_url.split(self.engine_type)[1]
            self._flowerpot = FlowerpotEngine.from_filename(file_path)

    @property
    def flowerpot(self) -> object:
        if self._flowerpot is None:
            self.connect()
        return self._flowerpot

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def query(self, **kwargs):
        """
        get a subset of :method:`data`.
        Allowed options-
        table : get rows from all files that start with that string followed by an underscore,
                followed by a number
        """
        table = kwargs.pop('table', None)
        if not table:
            raise NotImplementedError("Only 'table' has been implemented.")

        self.connect()
        # could use a regex
        selected_files = []
        for file in self._flowerpot.file_names:
            if not file.startswith(table):
                continue
            suffix = file[len(table):]
            if suffix[0] != '_':
                continue
            numeric_part = suffix[1:].split('.', 1)[0]
            if not numeric_part.isdecimal():
                continue
            selected_files.append(file)

        if len(selected_files) == 0:
            raise ValueError("Table doesn't exist")

        return self._flowerpot.items(file_name=selected_files)

    @property
    def data(self) -> Generator:
        self.connect()
        return self._flowerpot.items()

    @property
    def schema(self):
        return None



try:
    # pipenv install elasticsearch
    from elasticsearch import Elasticsearch
except ModuleNotFoundError:
    pass

from ayeaye.connectors.base import DataConnector, AccessMode


class ElasticsearchConnector(DataConnector):
    engine_type = "elasticsearch://"

    def __init__(self, *args, **kwargs):
        """
        Connector to Elasticsearch - https://www.elastic.co/elasticsearch/
        Args: @see :class:`connectors.base.DataConnector`

        This connector is very much a beta. Limitations and not fully thought through.
        e.g.
         - just one node

        additional args for ElasticsearchConnector
         None

        Connection information-
            engine_url format is elasticsearch://es_server[:port]/[<default_index>]
            e.g.
                elasticsearch://localhost/my_documents
        """
        super().__init__(*args, **kwargs)

        # set by :method:`connect`
        self.host = self.port = self.default_index = None
        self.client = None

    def close_connection(self):
        self.client = None

    def connect(self):
        if self.client is None:
            self.host, self.port, self.default_index = self._decode_engine_url()
            es_node_args = {"host": self.host}
            if self.port:
                es_node_args["port"] = self.port
            self.client = Elasticsearch([es_node_args])

    def _decode_engine_url(self):
        """
        Returns:
            host, port, default_index
            host, default_index are (str)
            port is (int)
        """
        r = dict(host=None, port=None, index=None)
        s_url = self.engine_url[len(self.__class__.engine_type) :]

        if "/" in s_url:
            host_port, r["index"] = s_url.split("/", 1)
            if r["index"] == "":
                r["index"] = None
        else:
            host_port = s_url

        if ":" in host_port:
            r["host"], r["port"] = host_port.split(":", 1)
            r["port"] = int(r["port"])
        else:
            r["host"] = host_port

        return r["host"], r["port"], r["index"]

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def add(self, document, doc_id=None, document_type=None, index=None):
        """
        Write document to index.

        @param document: (mixed) must be serialisable to JSON
        @param doc_id: document unique identifier. Will overwrite existing docs. with same id.
        @param document_type: (str)
        @param index: (str) must be set in engine_url or here
        @return: (dict) from ES library
        """
        # TODO expand data to include binary and instance of :class:`Pinnate` but needs a way of
        # de-serialising on retrieve.

        if self.access != AccessMode.WRITE and self.access != AccessMode.READWRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        self.connect()

        resolved_index = index or self.default_index
        if not resolved_index:
            raise ValueError("Unknown index: must be set in engine_url or as argument")

        # index(index, body, doc_type=None, id=None, params=None, headers=None)
        w = self.client.index(index=resolved_index, doc_type=document_type, id=doc_id, body=document)
        return w

    def fetch(self, doc_id=None, document_type=None, index=None):
        """
        @see :method:`add` but without `document` argument
        @return: the orginal document.
        """
        if self.access != AccessMode.READ and self.access != AccessMode.READWRITE:
            raise ValueError("Read attempted on dataset opened in WRITE mode.")

        self.connect()

        resolved_index = index or self.default_index
        if not resolved_index:
            raise ValueError("Unknown index: must be set in engine_url or as argument")

        # get(index, id, doc_type=None, params=None, headers=None)
        r = self.client.get(resolved_index, doc_id, doc_type=document_type)
        return r["_source"]



import copy
from enum import Enum

from ayeaye.connectors import connector_factory
from ayeaye.connectors.models_connector import ModelsConnector
from ayeaye.connectors.multi_connector import MultiConnector
from ayeaye.connectors.placeholder import PlaceholderDataConnector


class Connect:
    """
    Connect to a dataset or modelset.

    A dataset is a grouping of related data. What makes data 'related' will be down to the
    subject-domain. Datasets have a type - it could be a JSON document, a CSV file or a database.

    A modelset is one or more :class:`Model`(s).

    The main responsibility of :class:`Connect` is to provide a concrete subclass of
    :class:`DataConnector` or :class:`ModelConnector`. :class:`DataConnector` in turn provides
    access to operations on that dataset type. e.g. read, write etc.

    :class:`Connect` can be used standalone (see below) but is really designed to be used as a
    class variable in an :class:`ayeaye.Model`:

    ```
        class FavouriteColours(ayeaye.Model):
            favourite_colours = ayeaye.Connect(engine_url='csv://data/favourite_colours.csv')
    ```

    An instance of :class:`Connect` as a class variable in an :class:`ayeaye.Model` is a
    declaration of a model's use of a dataset. In Python, class variables shouldn't be used as
    'dynamic' instance variables so :class:`Connect` is a descriptor
    (https://docs.python.org/3/howto/descriptor.html) and the instantiation of the underlying
    connection (concrete subclass of :class:`DataConnector`) is evaluated on demand.

    It is also possible to use :class:`Connect` in *standalone* mode. This is as a convenience, in
    particular for evaluating datasets, for example in a Jupyter notepad:
    ```
    for row in ayeaye.Connect(engine_url='csv://data/favourite_colours.csv'):
      print(row.colour)
    ```

    For secrets management @see :class:`ConnectorResolver`.
    """

    mutually_exclusive_selectors = ["ref", "engine_url", "models"]

    class ConnectBind(Enum):
        MODEL = "MODEL"
        STANDALONE = "STANDALONE"
        NEW = "NEW"

    def __init__(self, **kwargs):
        """
        typical kwargs are 'ref', 'engine_url', 'access' TODO
        """
        self.base_constructor_kwargs = copy.copy(kwargs)
        self._construct(**kwargs)

    def _construct(self, **kwargs):
        """
        setup -- can be called either on true construction or when variables are overlaid on a
        (possibly cloned) instance of :class:`Connect`.
        """
        # :class:`Connect` is responsible for resolving 'ref' into an engine_url via a
        # data catalogue. So 'ref' isn't passed to data type specific connectors (i.e.
        # subclasses of :class:`DataConnector`)

        # these are passed to the data type specific connectors
        self.relayed_kwargs = {**self.base_constructor_kwargs, **kwargs}

        # check construction args are valid

        # mutually exclusive args
        a = [self.relayed_kwargs.get(s) is not None for s in self.mutually_exclusive_selectors]
        mandatory_args_count = sum(a)
        if mandatory_args_count > 1:
            raise ValueError("The kwargs ref, engine_url and models are mutually exclusive.")

        self.ref = self.relayed_kwargs.pop("ref", None)
        self._standalone_connection = None  # see :method:`data`
        self._parent_model = None

    def __repr__(self):
        args = ", ".join([f"{k}={v}" for k, v in self.base_constructor_kwargs.items()])
        return f"<Connect({args})>"

    def update(self, **kwargs):
        """
        Overlay (and overwrite) kwarg on existing instance.
        @returns None
        """
        # reminder to me: I've gone in circles here on returning a copy (factory style) and it's
        # not a good idea
        self._construct(**kwargs)

    def clone(self, **kwargs):
        """
        Overlay (and overwrite) kwarg onto a copy of the existing instance.

        This is typically used when :class:`Connect` objects are being used as class variables to
        refer to the same dataset multiple times in a single model.

        @see :method:`TestModels.test_double_usage`

        @return (instance of :class:`Connect`)
        """
        new_instance = copy.copy(self)
        new_instance._construct(**{**self.relayed_kwargs, **kwargs})
        return new_instance

    def connect_id(self):
        """
        Create an identity reference which is used when examining if separate Connect instances
        are actually referring to the same dataset/models.

        @return: (str)
        """
        for s in self.mutually_exclusive_selectors:
            if s in self.relayed_kwargs:
                # note, self.relayed_kwargs[s] could be a callable. Whatever it it needs to
                # deterministically cast to a string.
                return f"{s}:{self.relayed_kwargs[s]}"
        return "empty:"

    def __hash__(self):
        return hash(self.connect_id())

    def __eq__(self, other):
        if type(self) is type(other):
            return self.connect_id() == other.connect_id()
        return False

    def __copy__(self):
        c = self.__class__(**self.base_constructor_kwargs)
        c.relayed_kwargs = copy.copy(self.relayed_kwargs)
        return c

    def __get__(self, instance, instance_class):
        if instance is None:
            # class method called.
            # This means `self` is currently an attribute of the class (so NOT an instance
            # variable).
            #
            # Class variables are kind of like constants because they are shared between multiple
            # instances of the class and you shouldn't mutate them so just return self.
            return self

        if self.connection_bind == Connect.ConnectBind.STANDALONE:
            raise ValueError("Attempt to connect as a model when already initiated as standalone")

        ident = id(self)
        if ident not in instance._connections:

            instance._connections[ident] = self._prepare_connection()

            # a Connect belongs to zero or one ayeaye.Model. Link in both directions.
            instance._connections[ident]._parent_model = instance
            self._parent_model = instance

        return instance._connections[ident]

    def __set__(self, instance, new_connection):
        """
        Replace an instance of :class:`ayeaye.Model`'s :class:`ayeaye.Connect` with another
        instance of `Connect`.
        """
        if not isinstance(new_connection, self.__class__):
            my_class = self.__class__.__name__
            raise ValueError(f"Only {my_class} instances can be set")

        self.__init__(**new_connection.relayed_kwargs)
        ident = id(self)
        instance._connections[ident] = self._prepare_connection()

    def _prepare_connection(self):
        """
        Resolve everything apart from secrets needed to access the engine behind this dataset.

        @return: (instance subclass of :class:`DataConnector`) or (None) when resolve not yet
            possible.
        """
        if self.ref is not None:
            raise NotImplementedError(
                ("Sorry! Dataset discovery (looking up engine_url from ref) " "hasn't been written yet.")
            )

        engine_url = self.relayed_kwargs.get("engine_url")
        if callable(engine_url):
            engine_url = engine_url()

        if "models" in self.relayed_kwargs:
            # could be a callable but shouldn't be instantiated yet, ModelsConnector does that
            connector_cls = ModelsConnector

        elif engine_url is None:
            # engine_url not yet available
            connector_cls = PlaceholderDataConnector

        else:
            if isinstance(engine_url, list):
                # compile time list of engine_url strings
                # might be callable or a dict or set in the future
                connector_cls = MultiConnector
            else:
                connector_cls = connector_factory(engine_url)

        # Make an independent copy of relay_kwargs because these come from class variables
        # so could be resolved again under a different context.
        # TODO: when a list of callables is needed to populate MultiConnector for example then make
        # this recursive
        detached_kwargs = {}
        for k, v in self.relayed_kwargs.items():

            if k == "engine_url":
                # this might have been a callable above
                detached_kwargs[k] = copy.deepcopy(engine_url)

            if callable(v) and k != "models":

                if k in connector_cls.preserve_callables:
                    # reference to the callable is preserved for the `connector_cls` to call. This is
                    # needed when the `connector_cls` supplies arguments to the callable.
                    detached_kwargs[k] = v

                else:
                    # any kwarg arg could be a simple callable (i.e. it's called without any arguments)
                    # The callable isn't expected to have __deep_copy__ method. It's time to use the
                    # results of the callable so call it now. Note, the callable is left in place in
                    # `relay_kwargs`.
                    detached_kwargs[k] = v()
            else:
                detached_kwargs[k] = copy.deepcopy(v)

        connector = connector_cls(**detached_kwargs)
        connector._connect_instance = self
        return connector

    @property
    def connection_bind(self):
        """
        Raises a ValueError if an indeterminate state is found.

        @return: (ConnectBind)

        @see class's docstring.

        An instance of :class:`Connect` can be a class variable for an :class:`ayeaye.Model`
        or
        Standalone mode
        or
        Not yet determined
        """

        if self._parent_model is None and self._standalone_connection is None:
            return Connect.ConnectBind.NEW

        if self._standalone_connection is not None:
            return Connect.ConnectBind.STANDALONE

        if self._parent_model is not None:
            return Connect.ConnectBind.MODEL

        msg = (
            "Parent already attached and standalone connection is present. This"
            " shouldn't ever happen. Please let us know how it did!"
        )
        raise ValueError(msg)

    def connect_standalone(self):
        """
        Make a standalone Connect.

        Connect is normally used as part of an class:`ayeaye.Model` and will connect to the target
        dataset on demand. It's also possible to use Connect as a standalone instance which
        proxies to the target dataset's instance. The standalone version is also stood up on demand
        but it can be done explicitly with this method if Connect is short cutting the as-a-proxy
        incorrectly. See :method:`__getattr__`.
        """
        if self.connection_bind == Connect.ConnectBind.MODEL:
            raise ValueError("Attempt to connect as standalone when already bound to a model")

        if self.connection_bind == Connect.ConnectBind.NEW:
            self._standalone_connection = self._prepare_connection()

    def __getattr__(self, attr):
        """
        proxy through to subclass of :class:`DataConnector` when used as a standalone Connect
        (i.e. not a class variable on :class:`ayeaye.Model`).
        """
        if self.connection_bind == Connect.ConnectBind.MODEL:
            cls_name = self.__class__.__name__
            attrib_error_msg = f"'{cls_name}' object has no attribute '{attr}'"
            raise AttributeError(attrib_error_msg)

        # Short cut to proxy
        # avoid instantiating the target DataConnector if the attribute that is being accessed
        # is known to Connect.
        # TODO - maybe make a list of permitted non-proxy attribs because the target DataConnector
        # might do things to certain attribs on construction so the real version would differ.
        # e.g. ConnectorResolver with engine_urls
        if self.connection_bind == Connect.ConnectBind.NEW and attr in self.relayed_kwargs:
            return self.relayed_kwargs[attr]

        self.connect_standalone()
        return getattr(self._standalone_connection, attr)

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        """
        more intuitive use of the data behind this Connect. i.e. proxy to a DataConnector.
        e.g.
        ...
        for record in Connect(ref="my_dataset"):
            print(record)
        """
        self.connect_standalone()
        yield from self._standalone_connection



"""
Created on 12 Feb 2020

@author: si
"""
from ayeaye.connectors.base import AccessMode, DataConnector
from ayeaye.connectors import connector_factory


class MultiConnector(DataConnector):
    engine_type = None  # See docstring in constructor

    def __init__(self, *args, **kwargs):
        """
        Connector to multiple other Connectors.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for CsvConnector
         None

        Connection information-
            Not normally initiated with `engine_type://...` url but by giving a list
            of engine_urls which each connect to another subclass of DataConnector
        """
        # MultiConnector passes a copy of the kwargs to each child connector. It's not known
        # until after Connect which ones are permitted by the target DataConnector so the
        # exception wont be raised until usage.
        base_kwargs = {"engine_url": kwargs.pop("engine_url", None), "access": kwargs.pop("access", AccessMode.READ)}
        super().__init__(*args, **base_kwargs)

        self.connector_kwargs = kwargs
        self._child_data_connectors = None  # on connect
        self._child_dc_mapping = {}

    def connect(self):

        if len(self.engine_url) != len(set(self.engine_url)):
            msg = "engine_url contains duplicates. Not yet supported."
            raise NotImplementedError(msg)

        if self._child_data_connectors is None:
            if not isinstance(self.engine_url, list):
                raise ValueError("Expected a list of engine_urls in self.engine_url")

            self._child_data_connectors = []
            for idx, engine_url in enumerate(self.engine_url):
                connector_cls = connector_factory(engine_url)
                connector = connector_cls(engine_url=engine_url, access=self.access, **self.connector_kwargs)
                self._child_data_connectors.append(connector)
                # this is the unresolved engine_url
                self._child_dc_mapping[engine_url] = idx

        else:
            # see if self._child_data_connectors is stale. The happens if engine_urls has been
            # changed since the last connect(). It's easier to detect the change on read than it is
            # to override the behaviour of engine_urls.append(..) to connect on demand.
            # Also, can't just drop and rebuild as they could be in use
            already_seen = len(self._child_dc_mapping)
            for idx, engine_url in enumerate(self.engine_url):
                if engine_url not in self._child_dc_mapping:
                    if idx < already_seen:
                        msg = f"Can't remap after engine_url removed for: {engine_url}"
                        raise NotImplementedError(msg)
                    else:
                        # new engine_url
                        connector_cls = connector_factory(engine_url)
                        connector = connector_cls(engine_url=engine_url, access=self.access)
                        self._child_data_connectors.append(connector)
                        self._child_dc_mapping[engine_url] = idx

                elif self._child_dc_mapping[engine_url] != idx:
                    raise Exception("Please tell the AyeAye developers how this exception happens!")

    def close_connection(self):

        if self._child_data_connectors:
            for c in self._child_data_connectors:
                c.close_connection()

        self._child_data_connectors = None  # on connect
        self._child_dc_mapping = {}

    def add_engine_url(self, engine_url):
        """
        A convenience method for adding engine_urls at run time and returning the resolved
        connector (to the new engine_url) in one call.

        multi_connector.engine_url.append(...) could also be used.

        @param engine_url: (str) unresolved engine_url (i.e. could contain {params} to be resolved
                by :class:`ayeaye.connect_resolve.ConnectorResolver`

        @return: (subclass of :class:`DataConnector`)
        """
        self.engine_url.append(engine_url)
        self.connect()

        idx = self._child_dc_mapping[engine_url]
        connector = self._child_data_connectors[idx]
        return connector

    def __len__(self):
        """
        How many datasets is this MultiConnector holding
        """
        self.connect()
        return len(self._child_dc_mapping)

    def __getitem__(self, key):
        """
        @param key: (str) engine_url

        @return: (subclass of :class:`DataConnector`)
        """
        self.connect()
        idx = self._child_dc_mapping[key]
        connector = self._child_data_connectors[idx]
        return connector

    def __iter__(self):
        self.connect()
        for dc in self._child_data_connectors:
            yield dc

    @property
    def data(self):
        self.connect()
        return self._child_data_connectors



"""
Created on 14 Jan 2020

@author: si
"""
import copy
import csv
import os

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class CsvConnector(DataConnector):
    engine_type = "csv://"
    optional_args = {
        "field_names": None,
        "required_fields": None,
        "expected_fields": None,
        "alias_fields": None,
    }

    def __init__(self, *args, **kwargs):
        """
        Connector to Comma Separated Value (CSV) files.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for CsvConnector
            fieldnames (sequence, probably a list (str) - Field names for all rows in file.
                    Using this argument when reading forces the CSV module to treat the first line
                    of the file as data, not as a header. When used in write mode it is the order
                    of fields in the output csv file.

            required_fields (sequence, set or list (str)) - fields must be present in header. Other
                    fields may also exist. Raises ValueError if not.
                    Error to call this in write mode. A ValueError will be raised.

            expected_fields (sequence (str)) - fields must be exactly present in header in the
                    order given. Other fields may NOT also exist. Raises ValueError if not.
                    Error to call this in write mode. A ValueError will be raised.

            alias_fields (sequence of str) or (dict)) - if sequence (i.e. a list) it must be the
                    same length as the actual fields in the header. If a dictionary the key is the
                    original field name and the value is alias for the field.
                    Error to call this in write mode or with 'fieldnames' argument. A ValueError
                    will be raised.

        Connection information-
            engine_url format is
            csv://<filesystem absolute path>data_file.csv[;start=<line number>][;end=<line number>][;encoding=<character encoding>]
        e.g. csv:///data/my_project/all_the_data.csv
        """
        super().__init__(*args, **kwargs)

        # fieldnames are loaded from construction args or from field. This will be unified when
        # schemas are implemented. For now, keep track so loading fieldnames from file doesn't
        # make a :method:`_reset`
        self.base_field_names = copy.copy(self.field_names)

        self.delimiter = ","
        self._reset()

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError("Read+Write access not yet implemented")

    def _reset(self):
        self.file_handle = None
        self.csv = None
        self._encoding = None
        self._engine_params = None
        self.file_size = None
        self.approx_position = 0
        self.field_names = copy.copy(self.base_field_names)

    @property
    def engine_params(self):
        """
        @return: (Pinnate) with .file_path
                        and optional: .encoding .start and .end
        """
        if self._engine_params is None:
            self._engine_params = self.ignition._decode_filesystem_engine_url(
                self.engine_url, optional_args=["encoding", "start", "end"]
            )

            if "encoding" in self._engine_params:
                self._encoding = self.engine_params.encoding

            for typed_param in ["start", "end"]:
                if typed_param in self.engine_params:
                    self.engine_params[typed_param] = int(self.engine_params[typed_param])

            if "start" in self._engine_params or "end" in self._engine_params:
                raise NotImplementedError("TODO")

        return self._engine_params

    @property
    def encoding(self):
        """
        default encoding. 'sig' means don't include the unicode BOM
        """
        if self._encoding is None:
            ep = self.engine_params
            self._encoding = ep.encoding if "encoding" in ep else "utf-8-sig"
        return self._encoding

    def close_connection(self):
        if self.file_handle is not None:
            self.file_handle.close()
        self._reset()

    def connect(self):
        if self.csv is None:

            if self.access == AccessMode.READ:

                if self.field_names is not None and self.alias_fields is not None:
                    msg = "Can't set field_names and alias them. Just use 'field_names'"
                    raise ValueError(msg)

                if self.field_names is not None:
                    extra_args = {"fieldnames": self.field_names}
                else:
                    extra_args = {}

                self.file_handle = open(self.engine_params.file_path, "r", encoding=self.encoding)
                self.file_size = os.stat(self.engine_params.file_path).st_size

                self.csv = csv.DictReader(
                    self.file_handle,
                    delimiter=self.delimiter,
                    **extra_args,
                )
                self.field_names = self.csv.fieldnames

                if self.required_fields is not None:

                    required = set(self.required_fields)
                    field_names = set(self.field_names)

                    if not required.issubset(field_names):
                        missing_fields = ",".join(list(required - field_names))
                        msg = f"Missing required field(s): {missing_fields}"
                        raise ValueError(msg)

                if self.expected_fields is not None and self.expected_fields != self.field_names:
                    diff_s = set(self.expected_fields).symmetric_difference(set(self.field_names))
                    diff = ",".join(diff_s)
                    diff_count = len(diff_s)
                    msg = f"Expected fields does match fields found in file. There are {diff_count} difference(s):"
                    msg += f" [{diff}] expected: ["
                    msg += ",".join(self.expected_fields)
                    msg += "] but found: ["
                    msg += ",".join(self.field_names)
                    msg += "]"
                    raise ValueError(msg)

                if self.alias_fields is not None:

                    if isinstance(self.alias_fields, dict):
                        replace_fields = [self.alias_fields.get(f, f) for f in self.csv.fieldnames]
                        self.csv.fieldnames = self.field_names = replace_fields

                    elif not isinstance(self.alias_fields, list) or len(self.alias_fields) != len(self.csv.fieldnames):
                        msg = (
                            "Alias fields must be a dictionary or list with same number of "
                            "items as fields in the file"
                        )
                        raise ValueError(msg)

                    else:
                        self.csv.fieldnames = self.field_names = self.alias_fields

            elif self.access == AccessMode.WRITE:

                if (
                    self.required_fields is not None
                    or self.expected_fields is not None
                    or self.alias_fields is not None
                ):
                    msg = (
                        "The optional arguments: 'required_fields', 'expected_fields', 'alias_fields' "
                        "can't be used in WRITE mode."
                    )
                    raise ValueError(msg)

                # auto create directory
                file_dir = os.path.dirname(self.engine_params.file_path)
                if not os.path.exists(file_dir):
                    os.makedirs(file_dir)

                self.file_handle = open(self.engine_params.file_path, "w", newline="\n", encoding=self.encoding)
                self.csv = csv.DictWriter(
                    self.file_handle,
                    delimiter=self.delimiter,
                    fieldnames=self.field_names,
                )
                self.csv.writeheader()

            else:
                raise ValueError("Unknown access mode")

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        self.connect()
        for raw in self.csv:
            # OSError: telling position disabled by next() call so this for now
            # str(x) will slightly over count 'None'. None is given by DictReader when
            # trailing commas are omitted for optional fields at end of row.
            self.approx_position += len(self.delimiter.join([str(x) for x in raw.values()]))
            yield Pinnate(data=raw)

    @property
    def data(self):
        raise NotImplementedError("TODO")

    @property
    def progress(self):
        if self.access != AccessMode.READ or self.file_size is None or self.approx_position == 0:
            return None

        return self.approx_position / self.file_size

    def add(self, data):
        """
        Write line to CSV file.
        @param data: (dict or Pinnate)
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        # until schemas are implemented, first row determines fields
        if self.csv is None and self.field_names is None:
            if isinstance(data, dict):
                self.field_names = list(data.keys())
            elif isinstance(data, Pinnate):
                self.field_names = list(data.as_dict().keys())

        self.connect()

        if isinstance(data, dict):
            self.csv.writerow(data)
        elif isinstance(data, Pinnate):
            self.csv.writerow(data.as_dict())
        else:
            raise ValueError("data isn't an accepted type. Only (dict) or (Pinnate) are accepted.")


class TsvConnector(CsvConnector):
    """
    Tab separated values. See :class:`CsvConnector`
    """

    engine_type = "tsv://"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # see note in CsvConnector
        self.base_field_names = copy.copy(self.field_names)
        self.delimiter = "\t"



from inspect import isclass

from .base import DataConnector
from .bigquery import BigQueryConnector
from .csv_connector import CsvConnector, TsvConnector
from .elasticsearch_connector import ElasticsearchConnector
from .fake import FakeDataConnector
from .flowerpot import FlowerPotConnector
from .gcs_flowerpot import GcsFlowerpotConnector, FlowerpotEngine
from .json_connector import JsonConnector
from .kafka_connector import KafkaConnector
from .ndjson_connector import NdjsonConnector
from .parquet_connector import ParquetConnector
from .sqlalchemy_database import SqlAlchemyDatabaseConnector
from .uncooked_connector import UncookedConnector

class ConnectorPluginsRegistry:

    def __init__(self):
        self.registered_connectors = [] # list of classes, not instances - publicly readable
        self.reset()

    def reset(self):
        "set registered connectors to just those that are built in"
        self.registered_connectors = [BigQueryConnector, CsvConnector, FlowerPotConnector,
                                      FakeDataConnector, KafkaConnector, ParquetConnector,
                                      TsvConnector, SqlAlchemyDatabaseConnector, JsonConnector,
                                      NdjsonConnector, ElasticsearchConnector, UncookedConnector,
                                      ]

    def register_connector(self, connector_cls):
        """
        @param connector_cls ():
        """
        if not isclass(connector_cls) or not issubclass(connector_cls, DataConnector):
            msg = "'connector_cls' should be a class (not object) and subclass of DataConnector"
            raise TypeError(msg)

        # MAYBE - a mechanism to specify the position/priority of the class
        self.registered_connectors.append(connector_cls)


# global registry
connector_registry = ConnectorPluginsRegistry()

def connector_factory(engine_url):
    """
    return a subclass of DataConnector
    @param engine_url (str):
    """
    # TODO - these really need to be on demand and possibly outside this project
    engine_type = engine_url.split('://', 1)[0] + '://'
    for connector_cls in connector_registry.registered_connectors:
        if isinstance(connector_cls.engine_type, list):
            supported_engines = connector_cls.engine_type
        else:
            supported_engines = [connector_cls.engine_type]

        if engine_type in supported_engines:
            return connector_cls

    raise NotImplementedError(f"Unknown engine in url:{engine_url}")




from enum import Enum

from ayeaye.ignition import Ignition, EngineUrlCase, EngineUrlStatus


class AccessMode(Enum):
    READ = "r"
    WRITE = "w"
    READWRITE = "rw"


class BaseConnector:
    """
    Abstract class for factory instances from :class:`Connect`.
    """

    def __init__(self):
        # set when :class:`ayeaye.Connect` builds subclass instances
        self._connect_instance = None  # instance of :class:`ayeaye.Connect`
        self._parent_model = None  # instance of :class:`ayeaye.Model`

    @property
    def connect_instance(self):
        """
        Instances of subclasses of :class:`DataConnector` and :class:`ModelConnector` are usually
        built by :class:`ayeaye.Connect`. `connect_instance` is a reference to make it easy to
        tweak an existing connect on a model. See :method:`TestConnect.test_update_by_replacement`
        for an example.
        """
        return self._connect_instance


class DataConnector(BaseConnector):
    engine_type = None  # must be defined by subclasses

    # subclasses should specify their optional kwargs. Values in this dict are default values.
    optional_args = {}

    # list/set of arg names that are present in optional_args and when they are assigned with a
    # callable this shouldn't be called by :class:`ayeaye.Connect` but should intead be passed
    # as is to the target subclass of :class:`ayeaye.DataConnector`.
    preserve_callables = []

    # TODO - make it possible for internal variable name to not match kwarg name, e.g. schema -> self._schema
    # TODO - these aren't always optional, handling missing mandatory args here

    def __init__(self, engine_url=None, access=AccessMode.READ, **kwargs):
        """
        API to interact with AyeAye-compatible data sources
        :param engine_url (string): The file or URI to the location of the dataset
        :param access (AccessMode): Whether the dataset accessed through this Connector is for
                input or output
        **kwargs are any params needed by subclasses

        Note that subclasses must call this constructor and should 'pop' their arguments so that
        none are left unprocessed. Any which are specified in optional_args will be set as
        attributes by this super constructor.
        """
        super().__init__()
        self.access = access

        # engine_urls may need resolution of templated variables (typically secrets and paths). The
        # :class:`Ignition` module does this and makes URLs croxx-operating system compatible.
        self.ignition = Ignition(engine_url)

        if isinstance(engine_url, str):
            engine_type = [self.engine_type] if isinstance(self.engine_type, str) else self.engine_type
            if not any([engine_url.startswith(et) for et in engine_type]):
                raise ValueError("Engine type mismatch")

        # process optional arguments with their defaults
        for arg in self.__class__.optional_args:
            setattr(self, arg, kwargs.pop(arg, self.__class__.optional_args[arg]))

        # Subclasses should consume any kwargs before this constructor is invoked
        if len(kwargs) > 0:
            what_are_these = ", ".join(kwargs.keys())
            raise ValueError(f"Unexpected argument(s): '{what_are_these}'")

    def __del__(self):
        self.close_connection()

    @property
    def engine_url(self):
        """
        A fully resolved engine url contains everything needed to connect to the data source. This
        includes secrets like username and password.

        Don't store the result of this property when locking or similar. @see :class:`Ignition`.

        @return: (str) the fully resolved engine url.
        """
        status, e_url = self.ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
        if status == EngineUrlStatus.OK:
            return e_url
        elif status == EngineUrlStatus.NOT_AVAILABLE:
            raise ValueError("Engine URL not available")
        else:
            raise ValueError(f"Engine URL failed to resolve: {status}")

    def update(self, **kwargs):
        """
        Rebuild this (subclass of) DataConnector. It can only be used if self was built by
        :class:`ayeaye.Connect` because Connect determines the subclass based on the engine type.
        If the call used the same engine_type as `self` then this wouldn't be needed.

        The instance of :class:`Connect` that `self` is referencing is updated with `kwargs`. The
        current prepared DataConnect subclass (i.e. self) that was assigned to the parent model is
        thrown away because the new args probably alter how the dataset is connected to. It will
        re-initialise on demand.

        @param kwargs: anything you'd pass to :class:`Connect`
        @return: None
        """
        if self._connect_instance is None:
            raise ValueError("Connect instance not referenced. See :method:`update` for more")

        self._connect_instance.update(**kwargs)
        ident = id(self._connect_instance)
        if self._parent_model is not None:
            # remove old Connector, it will be re-built on access
            del self._parent_model._connections[ident]

    def connect(self):
        """
        Open resource handles used to access the dataset. e.g. network or filesystem connection.
        These resources are help open by the subclass.
        """
        pass

    def close_connection(self):
        """
        Explicitly close the connection to the dataset rather than just wait for the process to end.
        """
        pass

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        """
        more intuative use of the data behind each the DataConnector.
        e.g.
        ...
        for record in self.my_dataset:
            print(record)
        """
        self.connect()
        return self.data

    @property
    def data(self):
        """
        Return the entire dataset. The returned objects could be lazy evaluated. Use of this method
        will get ugly with large datasets. Better to use the :method:`__iter__` iterator.

        @return: mixed
        """
        raise NotImplementedError("TODO")

    def as_pandas(self):
        """
        Similar to :method:`data` but as a Pandas dataframe.

        @return: (Pandas dataframe)
        """
        raise NotImplementedError("Not available for all datasets or might need to be written")

    @property
    def progress(self):
        """
        Return a number between and including 0 to 1 as simplified representation of how much of
        the dataset has been read.

        Optional property that can be implemented by subclasses. When available it returns float
        from 0 -> 1 describing the position within the dataset. The value is only meaningful when
        the dataset is in READ mode and records are being iterated through. The value isn't
        expected to be really accurate as some reads will cache data so exact file positions wont
        be available.

        :returns: (float) or None when not available.
        """
        return None

    @property
    def datasource_exists(self):
        msg = (
            "Can optionally be implemented by subclasses. "
            "Contribute if you need it! There is an example in JsonConnector."
        )
        raise NotImplementedError(msg)



from collections import namedtuple
from datetime import datetime
from enum import Enum
import os
from time import time
import warnings

import ayeaye
from ayeaye.connectors.base import DataConnector
from ayeaye.connect_resolve import connector_resolver
from ayeaye.executors.multiprocess import ProcessPool
from ayeaye.ignition import EngineUrlCase, EngineUrlStatus


class LockingMode(Enum):
    """
    How to capture context around datasets in a model. This is used to track data provenance and
    to make model repeatability possible.

    @see :method:`Model.lock`
    """

    CONTEXT = "context"  # just the ayeaye.connector_resolver context
    ALL_DATASETS = "all_datasets"  # the engine_urls for all datasets


class Model:
    """
    Do the thing!

    Abstract class

    The thing is probably an ETL (Extract, Transform and Load) task which is at the minimum
    the :method:`build`. It could optionally also have a :method:`pre_build_check`, which is run
    first and must succeed. After the :method:`build` an optional :method:`post_build_check`
    can be used to see if build worked.
    """

    def __init__(self):
        self._connections = {}  # see :class:`ayeaye.Connect` 'descriptors' in doc string.

        self.log_to_stdout = True
        self.external_logger = None
        self.progress_log_interval = 20  # minimum seconds between progress messages to log

        # stats
        self.start_build_time = None
        self.progress_logged = None  # time last logged

    def go(self):
        """
        Run the model.

        The steps to run a model are
        1. :method:`pre_build_check` - Optional conditional check if model's pre-conditions have
           been met.
        2. :method:`build` - The main modelling stage
        3. :method:`post_build_check` - Optional check if the outputs are valid. e.g. a data
           validation check that is simple and concise. It's like a unit test but on changable
           data.

        Datasets are closed after each stage. This ensures the connections don't have a state, for
        example a position within a file.

        @return: boolean for success
        """
        self.start_build_time = time()
        if not self.pre_build_check():
            self.log("Pre-build check failed", "ERROR")
            self.close_datasets()
            return False
        self.close_datasets()

        self._build()
        self.close_datasets()

        if not self.post_build_check():
            self.log("Post-build check failed", "ERROR")
            self.close_datasets()
            return False
        self.close_datasets()

        return True

    def pre_build_check(self):
        """
        Optionally implemented by subclasses to check any conditions that must be met before
        running :method:`build`.

        For example, assumptions :method:`build` makes on on the format or values within the data
        could be checked here in order to keep code in build simple.

        @return: boolean. If False is returned :method:`build` won't be run.
        """
        return True

    def _build(self):
        """
        Internal stub to allow :class:`Model` and :class:`PartitionedModel` to do different things.
        """
        self.build()

    def build(self):
        """
        Do the thing! - build/process/transform - whatever the model does starts here.

        Must be implemented by subclasses. Don't change method argument in subclass as this is
        called by :method:`go` when running the full model execution.
        """
        raise NotImplementedError("All models must implement the build() method")

    def post_build_check(self):
        """
        This is an optional method that will be run after :method:`build`. It can be used to check
        the validity of the build process.

        @return: boolean. If False is returned the model is considered to have failed.
        """
        return True

    @classmethod
    def connects(cls):
        """
        :returns (dict) of :class:`Connect` classes declared as class variables for this model.
                key is class variable name
                value is :class:`ayeaye.Connect`
        """
        # find :class:`ayeaye.Connect` connections to datasets
        connects = {}
        for obj_name in dir(cls):
            obj = getattr(cls, obj_name)
            if isinstance(obj, ayeaye.Connect):
                connects[obj_name] = obj

        return connects

    def datasets(self):
        """
        :returns (dict) of dataset connections for this model.
                key is class variable name
                value is :class:`ayeaye.DataConnector`
        """
        # find :class:`ayeaye.Connect` connections to datasets
        connections = {}
        for obj_name in dir(self):
            obj = getattr(self, obj_name)
            if isinstance(obj, DataConnector):
                connections[obj_name] = obj

        return connections

    def close_datasets(self):
        """
        Call :method:`close_connection` on all datasets.
        """
        for connection in self.datasets().values():
            connection.close_connection()

    def set_logger(self, logger):
        """
        Also log to something with a :method:`write`.
        e.g. StringIO
        """
        self.external_logger = logger

    def log(self, msg, level="INFO"):
        """
        @param level: (str) DEBUG, PROGRESS, INFO, WARNING, ERROR or CRITICAL
        """
        if not (self.log_to_stdout or self.external_logger is not None):
            return

        date_str = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
        msg = "{} {}{}".format(date_str, level.ljust(10), msg)

        if self.external_logger is not None:
            self.external_logger.write(msg)

        if self.log_to_stdout:
            print(msg)

    def log_progress(self, position_pc, msg=None):
        """
        Externally provided indication of position through build is used to calculate the remaining
        time until this model completes. A `PROGRESS` log message is issued a maximum of every
        `self.progress_log_interval` seconds (default is 20 seconds) so this method can be called
        more frequently and the output log wont be swamped.

        @param position_pc: (float) or None meaning ignore - between 0.0 and 1.0 (complete)
        @param msg: (str) additional user friendly info
        """
        time_now = time()

        if position_pc > 0.0001 and (
            self.progress_logged is None or self.progress_logged + self.progress_log_interval < time_now
        ):

            if msg is None:
                msg = ""

            progress_pc = "{:.2f}%".format(position_pc * 100)
            running_secs = time_now - self.start_build_time
            eta = "{:.2f}".format((running_secs / position_pc) - running_secs)
            self.log(f"{progress_pc} {eta} seconds remaining. {msg}", level="PROGRESS")
            self.progress_logged = time_now

    def fetch_locking(self):
        """
        A 'lock' is the information needed by a model to reproduce an output.

        This method is optionally implemented by subclasses that need more than the context
        provided by `ayeaye.connector_resolver` in order to be repeatable. For example, the model
        might not be deterministic.

        The value returned will be passed to :method:`apply_locking` if/when a model needs to
        be repeated.

        @returns something that can be serialised to JSON
        """
        return None

    def apply_locking(self, lock_doc):
        """
        Use the output from :method:`fetch_locking` to reproduce results from a previous build.

        This method should be implemented if a model needs to be hydrated with locking details from
        a previous build.
        """
        raise NotImplementedError("Missing sub-class method on attempt to re-hydrate locking.")

    def lock(self, locking_level=LockingMode.CONTEXT):
        """
        A 'lock' is the information needed by a model to reproduce an output.

        Return a JSON safe dictionary of variables needed to repeat the build.

        The returned dictionary is expected to be serialised and stored.

        @param locking_level: (LockingMode)
            LockingMode.CONTEXT - key values from ayeaye.connector_resolver
            LockingMode.ALL_DATASETS - CONTEXT + engine_urls from all datasets in model.

        @return (dict)
        """
        # this is very much work in progress.
        # coming soon
        # - code version (git commitishes)
        # - library code versions (maybe pipenv lock file)

        locking_doc = {"resolve_context": connector_resolver.capture_context()}

        model_lock = self.fetch_locking()
        if model_lock is not None:
            locking_doc["model_locking"] = model_lock

        if locking_level == LockingMode.ALL_DATASETS:
            locking_doc["dataset_engine_urls"] = {}
            for dataset_name, connector in self.datasets().items():

                # TODO: this should be EngineUrlCase.WITHOUT_SECRETS. Secrets shouldn't be in
                # locking doc. But that's not implemented yet,
                status, engine_url = connector.ignition.engine_url_at_state(EngineUrlCase.FULLY_RESOLVED)
                msg = (
                    "Incomplete implementation: if there are secrets in the engine_url they "
                    "will be included in locking doc."
                )
                warnings.warn(msg)

                if status != EngineUrlStatus.OK:
                    raise ValueError(f"Can't lock, engine_url not available for '{dataset_name}'")

                locking_doc["dataset_engine_urls"][dataset_name] = engine_url

        return locking_doc


class PartitionedModel(Model):
    """
    Similar to :class:`Model` but requires additional methods to be implemented by the subclass to
    describe how the processing can be split into parallel subtasks.

    A model can suggest to the executor how many parallel tasks could be used and the executor
    is responsible for orchestrating a level of parallelisation that fits the execution environment.

    An external executor is expected to run :class:`PartitionedModel` subclasses. It would
    instantiate the model and examine it's :method:`partition_plea` before running it in a
    distributed environment (e.g. Google Cloud Run, Celery, AWS' ECS etc.). The external executor
    is responsible for transporting task arguments, log message etc. TODO : there isn't an open
    source executor yet. It has been done in a commercial project.

    :class:`PartitionedModel` does support a simple mechanism for parallel execution across local
    processes. To use it implement the mandatory methods and run method:`go`.

    e.g.
    >>> my_model = MyModel()
    >>> my_model.go()

    And a reasonable number of local processes will run in a Python multiprocessor Pool to work
    on the tasks the model divides itself into.

    The executor is responsible for re-executing failed tasks. All sub-tasks should be idempotent and
    therefore safe to execute multiple times, possibly in parallel.

    It works as follows-

    1. The executor creates the parent instance. If the 'simple mechanism for parallel execution' (as
    above) is being used, this is the instance that :method:`go` is running on.

    2. The executor asks the parent instance for 'suggestions' of how many sub-tasks the model can
    be split into (see :method:`partition_plea`). This suggestion is for the number of workers the
    model would like to have running in parallel. There are situations where an exact number of
    workers are needed (e.g. so each worker has independent control of a resource such as a file)
    and there are other situations where too many worker could result in a stampede on a resource
    (such as a database).

    3. The executor evaluates the :class:`PartitionOption` returned by :method:`partition_plea`
    along with the capabilities of the execution environment and requests a mutually compatible
    number of partition arguments from :method:`partition_slice`. The number of sub-tasks doesn't
    need to match the number of workers although there can be a relationship.

    4. :method:`partition_slice` is called on the parent instance. It returns a list of method
    names and sub-task arguments. Each of which is passed by the executor to an instance of the
    model that has been instantiated with the same resolver context (see :class:`ConnectorResolver`)
    as the 'parent' model. The models running in the worker processes are initiated when the worker
    process starts, i.e. each instance will have it's sub-task method called multiple times. See
    :method:`partition_initialise`.

    4. The return value from each subtask is passed to optional :method:`partition_subtask_complete`.

    5. When the executor is satisfied that all sub-tasks are complete the optional
    :method:`partition_complete` method is called.

    6. The :method:`build` will be called in a separate worker process. i.e. this isn't the same
    instance as the parent instance.
    """

    # Start simple, this will no doubt increase in flexibility. Each are an integer suggesting
    # how many sub-tasks the execution could be split into.
    PartitionOption = namedtuple("PartitionOption", ("minimum", "maximum", "optimal"))

    def partition_initialise(self):
        """
        This method will be called when a worker process instantiates a model. This method can
        be overridden but sub-class's method must be called.
        """
        self.start_build_time = time()

        return None

    def partition_plea(self):
        """
        Subclass to suggest possible options for splitting the task execution.

        This is the mechanism for the model to explain to the execution environment how the task
        could be split into sub-tasks.

        The default behaviour is to return `PartitionOption(minimum=1, maximum=128, optimal=16)`.

        @return (PartitionOption)
        """
        return PartitionedModel.PartitionOption(minimum=1, maximum=128, optimal=16)

    def partition_slice(self, partition_count):
        """
        Create sub-task arguments.

        Create a list (sorry, no iterators) of arguments to define each subtask. This is composed
        of the method name and key word arguments. The method+kwargs will be called on one of the
        workers.

        @returns (list of (method name (str), kwargs [i.e. a dict])
        """
        raise NotImplementedError("All models must implement this method")

    def partition_subtask_complete(self, subtask_method_name, subtask_kwargs, subtask_return_value):
        """
        Optional method. Takes return values from subtasks.

        This will be called on the parent task (i.e. not on the worker). It can be used to collate
        results or take further actions when a sub-task has finished.
        """
        return None

    def partition_complete(self):
        """
        Optional method. called when executor has finished all sub-tasks.
        """
        return None

    def _build(self):
        """
        When not run by an executor in a distributed environment the default behaviour is for
        :class:`PartitionedModel`s to behave like normal :class:`Model`s but using
        multiprocessing for local parallel execution.

        This 'built in' concurrent operator sets up a pool of processes that is related to the
        number of sub-tasks indicated by :method:`partition_plea`. There doesn't need to be
        a relationship between the number of processes and the number of tasks but it wouldn't
        make sense for more processes than tasks.
        """

        partition_option = self.partition_plea()

        assert partition_option.minimum > 0
        assert partition_option.minimum <= partition_option.maximum

        workers_count = partition_option.minimum

        # TODO - make this user adjustable
        typical_cpu_bounding_ratio = 2
        max_recommended_workers = os.cpu_count() * typical_cpu_bounding_ratio

        if partition_option.optimal < max_recommended_workers:
            workers_count = partition_option.optimal
        else:
            workers_count = max_recommended_workers

        if workers_count > partition_option.maximum:
            workers_count = partition_option.maximum

        # workers_count =  1
        self.log(f"Using {workers_count} worker processes")

        tasks = [("build", None)]  # all subclasses of :class:`Model` must include `build` method
        tasks.extend(self.partition_slice(workers_count))

        proc_pool = ProcessPool(processes=workers_count)
        for subtask_return in proc_pool.run_subtasks(model_cls=self.__class__, tasks=tasks):

            method_name, original_method_kwargs, subtask_return_value = subtask_return
            self.partition_subtask_complete(
                subtask_method_name=method_name,
                subtask_kwargs=original_method_kwargs,
                subtask_return_value=subtask_return_value,
            )

        self.partition_complete()



import json

class Pinnate:
    """
    Dictionary or attribute access to variables loaded either from a JSON
    string or supplied as a dictionary.

    >>> a = Pinnate({'my_string':'abcdef'})
    >>> a.my_string
    'abcdef'
    >>> a['my_string']
    'abcdef'
    >>> a.as_dict()
    {'my_string': 'abcdef'}

    objects within lists-

    >>> from ayeaye.pinnate import Pinnate
    >>> d={'my_things' : [1,2,{'three':3}]}
    >>> a = Pinnate(d)
    >>> a.my_things
    [1, 2, <ayeaye.pinnate.Pinnate object at 0x108526e10>]
    >>> a.my_things[2].three
    3
    """
    def __init__(self, data=None):
        """
        :param data: dictionary or dictionary encoded in json or instance of Pinnate
        """
        self._attr = {}
        if isinstance(data, self.__class__):
            self._attr = data._attr
        elif data:
            self.load(data, merge=False)

    def __unicode__(self):
        d = ', '.join([u"{}:{}".format(k, v) for k, v in self._attr.items()])
        return '<Pinnate %s>' % d

    def __str__(self):
        return self.__unicode__().encode("ascii", "replace").decode()

    def keys(self):
        return self._attr.keys()

    def values(self):
        return self._attr.values()

    def items(self):
        return self._attr.items()

    def __contains__(self, key):
        return key in self._attr

    def as_dict(self, select_fields=None):
        """
        @param select_fields: (list of str) to only include some fields from model.
        @return: (dict) with mixed values
        """
        if select_fields is not None:
            r = {}
            for k in select_fields:
                if isinstance(self._attr[k], self.__class__):
                    v = self._attr[k].as_dict()
                else:
                    v = self._attr[k]
                r[k] = v
            return r
        else:
            return {k: v.as_dict() if isinstance(v, self.__class__) else v \
                    for k, v in self._attr.items()}

    def as_json(self, *args, **kwargs):
        """
        @see :method:`as_dict` for params.
        @returns (str) JSON representation
        """
        return json.dumps(self.as_dict(*args, **kwargs), default=str)

    def __getattr__(self, attr):
        if attr not in self._attr:
            raise AttributeError("{} instance has no attribute '{}'".format(self.__class__.__name__, attr))
        if isinstance(self._attr[attr], list):

            def list_recurse(item):
                r = []
                for s in item:
                    if isinstance(s, dict):
                        r.append(self.__class__(s))
                    elif isinstance(s, list):
                        r.append(list_recurse(s))
                    else:
                        r.append(s)
                return r

            return list_recurse(self._attr[attr])

        elif isinstance(self._attr[attr], dict):
            return self.__class__(self._attr[attr])
        else:
            return self._attr[attr]

    def __setattr__(self, attr, val):
        super(Pinnate, self).__setattr__(attr, val)
        if attr != '_attr':
            self._attr[attr] = val

    def __getitem__(self, key):
        return self._attr[key]

    def __setitem__(self, key, value):
        self._attr[key] = value

    def get(self, key, default=None):
        return self._attr.get(key, default)

    def load(self, data, merge=False):
        """
        :param data: dict or json string
        :param merge: bool see :method:`update` if False or :method:`merge` when True.
        """

        if not isinstance(data, dict):
            data = json.loads(data)

        if merge:
            self.merge(data)
        else:
            self.update(data)

    def update(self, data):
        """
        Extend the Pinnate with further settings. If a setting with an existing key is supplied,
        then the previous value is overwritten.

        :param data: dictionary or dictionary encoded in json
        """
        for k, v in data.items():
            if isinstance(v, dict):
                self._attr[k] = Pinnate(v)
            else:
                self._attr[k] = v

    def merge(self, data):
        """
        Extend the Pinnate with further settings. If a setting with an existing key is supplied,
        then the previous value is either updated (if the previous value is a dict) or overwritten
        (if the previous value is a scalar).

        Where corresponding values are not of compatible types, a ValueError is raised. Compatible
        means that an existing dict value must remain a dict (thus the dicts are merged), or a
        non-dict value must remain a non-dict type.

        :param data: dictionary or dictionary encoded in json
        """
        for k, v in data.items():
            if isinstance(v, dict):
                try:
                    self._attr[k].merge(v)
                except KeyError:
                    self._attr[k] = Pinnate(v)
                except AttributeError:
                    raise ValueError("Invalid key '{}'".format(k))
            else:
                if k in self._attr and isinstance(self._attr[k], self.__class__):
                    msg = ("Key '{}' attempted to overwrite an existing Pinnate."
                            "Operation not permitted."
                            )
                    raise ValueError(msg.format(k))
                self._attr[k] = v



"""
Created on 17 Dec 2020

@author: si
"""
import os

try:
    import ndjson
except ModuleNotFoundError:
    pass

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class NdjsonConnector(DataConnector):
    engine_type = "ndjson://"

    def __init__(self, *args, **kwargs):
        """
        Connector to Newline Delimited JSON files (http://ndjson.org/).

        ndjson is one json document per line.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for NdjsonConnector
         None

        Connection information-
            engine_url format is
            ndjson://<filesystem absolute path>[;encoding=<character encoding>]
        e.g. ndjson:///data/my_project/the_data.json;encoding=latin-1
        """
        super().__init__(*args, **kwargs)

        self._reset()

        if self.access == AccessMode.READWRITE:
            raise NotImplementedError("READWRITE access not yet implemented")

    def _reset(self):
        self.file_handle = None
        self.reader = None
        self.writer = None
        self._encoding = None
        self._engine_params = None
        self.file_size = None
        self.approx_position = 0

    @property
    def engine_params(self):
        """
        @return: (Pinnate) with .file_path
                        and optional: .encoding
        """
        if self._engine_params is None:
            self._engine_params = self.ignition._decode_filesystem_engine_url(
                self.engine_url, optional_args=["encoding"]
            )

            if "encoding" in self._engine_params:
                self._encoding = self.engine_params.encoding

        return self._engine_params

    @property
    def encoding(self):
        """
        default encoding. 'sig' means don't include the unicode BOM
        """
        if self._encoding is None:
            ep = self.engine_params
            self._encoding = ep.encoding if "encoding" in ep else "utf-8-sig"

        return self._encoding

    def close_connection(self):
        if self.file_handle is not None:
            self.file_handle.close()
        self._reset()

    def connect(self):
        if self.reader is None and self.writer is None:

            if self.access == AccessMode.READ:
                self.file_handle = open(self.engine_params.file_path, "r", encoding=self.encoding)
                self.file_size = os.stat(self.engine_params.file_path).st_size
                self.reader = ndjson.reader(self.file_handle)

            elif self.access == AccessMode.WRITE:
                self.file_handle = open(self.engine_params.file_path, "w", encoding=self.encoding)
                self.writer = ndjson.writer(self.file_handle)

            else:
                raise ValueError("Unknown access mode")

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        self.connect()

        for r in self.reader:
            # OSError: telling position disabled by next() call so this for now.
            # TODO: It's a waste of CPU to make it back into a string.
            self.approx_position += len(str(r))
            yield Pinnate(data=r)

    @property
    def data(self):
        raise NotImplementedError("TODO")

    @property
    def progress(self):
        if self.access != AccessMode.READ or self.file_size is None or self.approx_position == 0:
            return None

        return self.approx_position / self.file_size

    def add(self, data):
        """
        Write record to ndjson file.
        @param data: (dict or Pinnate) - must be safe to serialise to JSON so no dates etc.
        """
        if self.access != AccessMode.WRITE:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        self.connect()

        if isinstance(data, dict):
            self.writer.writerow(data)
        elif isinstance(data, Pinnate):
            self.writer.writerow(data.as_dict())
        else:
            raise ValueError("data isn't an accepted type. Only (dict) or (Pinnate) are accepted.")



import copy
from inspect import isclass

from ayeaye.connectors.base import AccessMode, BaseConnector
from ayeaye.model import Model
from ayeaye.pinnate import Pinnate


class ModelsConnector(BaseConnector):
    def __init__(self, models):
        """
        Connector to run :class:`ayeaye.Models`.

        For args: @see :class:`connectors.base.BaseConnector`

        additional args for ModelsConnector
         models (class, list, set or callable) all of which are :class:`ayeaye.Models`. They aren't
                    instances of the class but the class itself. A *list* is an explicit ordering
                    that the models should be run in. Models in *set* will be examined to see
                    dataset dependencies and these will be used to determine the model run order.

        """
        super().__init__()

        invalid_construction_msg = (
            "models must be a class, list, set or callable. All of which "
            "result in one or more :class:`ayeaye.Models` classes (not "
            "instances)."
        )

        # validate and prepare
        if isclass(models) and issubclass(models, Model):
            models = [
                models,
            ]

        if callable(models):
            # TODO
            raise NotImplementedError("TODO")

        elif isinstance(models, (list, set)):
            if not all([isclass(m) and issubclass(m, Model) for m in models]):
                raise ValueError(invalid_construction_msg)
            self.models = models

        else:
            raise ValueError(invalid_construction_msg)

    def __len__(self):
        return len(self.models)

    def __iter__(self):
        """
        yield model classes. No ordering when models are a set but does honour order when models
        is a list
        """
        yield from self.models

    def _resolve_run_order(self):
        """
        Use the dataset connections in each model to determine the dependencies between models and
        therefore the order to run them in.

        @return: (Pinnate) with attributes:
                    leaf_sources (set of datasets) - read but not written to
                    leaf_targets (set of datasets) - written to but not read - end goal of model
                    run_order (list of sets of nodes) - All models in a set can be run in parallel.
                            Each set must be complete before the next set is run.

                    'datasets' as subclasses of :class:`Connect`
                    `nodes` are type :class:`Pinnate` with attributes model_cls, model_name,
                        targets and sources
        """
        # This algorithm is a bit overly simplistic and sub-optimal. The return is a list of sets
        # with all models in set needing to be run. But the set is actually the models that were
        # ready to run when the prior models are complete. So there could be subsequent models that
        # are only waiting on a subset of each set.

        # Find all the datasets in all the model and classify datasets as 'sources' (READ access)
        # and 'targets' (WRITE access) or READWRITE for both.
        all_targets = set()
        all_sources = set()
        nodes = {}
        for model_cls in self.models:
            # TODO find ModelConnectors and recurse into those
            model_name = model_cls.__name__
            if model_name in nodes:
                raise ValueError(f"Duplicate node found: {model_name}")

            node = Pinnate({"model_cls": model_cls, "model_name": model_name, "targets": set(), "sources": set()})
            for dataset in model_cls.connects().values():
                if dataset.access in [AccessMode.READ, AccessMode.READWRITE]:
                    node.sources.add(dataset)
                    all_sources.add(dataset)

                if dataset.access in [AccessMode.WRITE, AccessMode.WRITE]:
                    node.targets.add(dataset)
                    all_targets.add(dataset)

            nodes[model_name] = node

        leaf_sources = all_sources - all_targets
        leaf_targets = all_targets - all_sources

        completed = copy.copy(leaf_sources)
        run_order = []
        while len(nodes) > 0:
            loop_ready = set()
            for node in nodes.values():
                if node.sources.issubset(completed):
                    loop_ready.add(node)

            if len(loop_ready) == 0:
                msg = "The set of models can't be built into a single acyclic graph."
                raise ValueError(msg)

            # nodes in loop_ready have been visited add their edges to the completed map
            run_order.append(loop_ready)
            for ready_node in loop_ready:
                node = nodes[ready_node.model_name]
                completed = completed.union(node.targets)
                del nodes[ready_node.model_name]

        p = Pinnate({"leaf_sources": leaf_sources, "leaf_targets": leaf_targets, "run_order": run_order})
        return p

    def run_order(self):
        """
        @return: (list) items in the list are either-
                    (i) sets of subclasses of :class:`ayeaye.Model`
                    OR
                    (ii) lists of sets as per (i)

        All models in a set can be run in parallel. Each set must be complete before the next set
        in the list is run.
        """
        if isinstance(self.models, list):
            # models in 'list' mode are wrapped as single item sets because all items in a set must
            # complete before next item in a list is run.
            return [{m} for m in self]

        elif isinstance(self.models, set):
            # _resolve_run_order returns leaf nodes and the run order is built using Pinnate
            # instances as nodes. Remove all this with to return a simple run order.
            resolved_order = self._resolve_run_order()

            def simplify_nodes(n):
                if isinstance(n, Pinnate):
                    return n.model_cls
                if isinstance(n, (set, list)):
                    rx = [simplify_nodes(nx) for nx in n]
                    if isinstance(n, set):
                        return set(rx)
                    return rx

            return [simplify_nodes(r) for r in resolved_order.run_order]

        else:
            raise ValueError("Unknown models container.")



"""
Created on 15 Apr 2020

@author: si
"""
import json
import os

from ayeaye.connectors.base import DataConnector, AccessMode
from ayeaye.pinnate import Pinnate


class JsonConnector(DataConnector):
    engine_type = "json://"

    def __init__(self, *args, **kwargs):
        """
        Single JSON file loaded into memory and made available as a :class:`Pinnate` object.

        For args: @see :class:`connectors.base.DataConnector`

        additional args for JsonConnector
         None

        Connection information-
            engine_url format is
            json://<filesystem absolute path>[;encoding=<character encoding>][;indent=<spaces when pretty printing write output>]
        e.g. json:///data/my_project/the_data.json;encoding=latin-1

        """
        super().__init__(*args, **kwargs)

        self._doc = None
        self._encoding = None
        self._engine_params = None

    @property
    def engine_params(self):
        if self._engine_params is None:
            self._engine_params = self.ignition._decode_filesystem_engine_url(
                self.engine_url, optional_args=["encoding", "indent"]
            )

            if "encoding" in self._engine_params:
                self._encoding = self.engine_params.encoding

            for typed_param in ["indent"]:
                if typed_param in self.engine_params:
                    self.engine_params[typed_param] = int(self.engine_params[typed_param])

        return self._engine_params

    @property
    def encoding(self):
        """
        default encoding. 'sig' means don't include the unicode BOM
        """
        if self._encoding is None:
            ep = self.engine_params
            self._encoding = ep.encoding if "encoding" in ep else "utf-8-sig"
        return self._encoding

    def close_connection(self):
        self._doc = None

    def connect(self):
        """
        When in AccessMode.READ, read the contents of the file into a :class:`Pinnate` instance
        that is available as self.data.

        In AccessMode.WRITE mode connect() doesn't do anything because file handles aren't kept
        open by the JsonConnector. The write operation is in :method:`_data_write`.
        """
        if self._doc is None:

            file_path = self.engine_params.file_path
            if (
                self.access in [AccessMode.READ, AccessMode.READWRITE]
                and os.path.isfile(file_path)
                and os.access(file_path, os.R_OK)
            ):

                with open(file_path, "r", encoding=self.encoding) as f:
                    as_native = json.load(f)
                    self._doc = Pinnate(as_native)

            else:
                raise ValueError(f"Attempt to read '{file_path}' which isn't readable")

    def __len__(self):
        raise NotImplementedError("TODO")

    def __getitem__(self, key):
        raise NotImplementedError("TODO")

    def __iter__(self):
        raise NotImplementedError("Not an iterative dataset. Use .data instead.")

    def _data_read(self):
        self.connect()
        return self._doc

    def _data_write(self, new_data):
        """
        Set the contents of a JSON file. `new_data` can be an instance of :class:`Pinnate` or any
        python datatype that will serialise into JSON.

        Will raise TypeError if the data can't be serialised to JSON.

        @param new_data: (mixed, see description)
        """
        if self.access not in [AccessMode.WRITE, AccessMode.READWRITE]:
            raise ValueError("Write attempted on dataset opened in READ mode.")

        json_args = {}
        if "indent" in self.engine_params:
            json_args["indent"] = self.engine_params["indent"]

        if isinstance(new_data, Pinnate):
            as_json = json.dumps(new_data.as_dict(), **json_args)
        else:
            as_json = json.dumps(new_data, **json_args)

        # Data is written to disk immediately. The file handle isn't left open.
        # @see :method:`connect`.
        with open(self.engine_params.file_path, "w", encoding=self.encoding) as f:
            f.write(as_json)

    data = property(fget=_data_read, fset=_data_write)

    @property
    def datasource_exists(self):
        """
        Returns:
            (bool) if the datasource referred to in self.engine_url exists.
        """
        return os.path.exists(self.engine_params.file_path)
